@@ -1,16 +1,6 @@
-extern crate proc_macro; // Needed even though it's the 2018 edition. See https://github.com/idanarye/rust-typed-builder/issues/57.
+#![no_std]
 
-use proc_macro2::TokenStream;
-
-use syn::parse::Error;
-use syn::spanned::Spanned;
-use syn::{parse_macro_input, DeriveInput};
-
-use quote::quote;
-
-mod field_info;
-mod struct_info;
-mod util;
+use core::ops::FnOnce;
 
 /// `TypedBuilder` is not a real type - deriving it will generate a `::builder()` method on your
 /// struct that will return a compile-time checked builder. Set the fields using setters with the
@@ -62,9 +52,9 @@ mod util;
 /// // Foo::builder().x(1).y(2).y(3);
 /// ```
 ///
-/// # Customisation with attributes
+/// # Customization with attributes
 ///
-/// In addition to putting `#[derive(TypedBuilder)]` on a type, you can specify a `#[builder(…)]`
+/// In addition to putting `#[derive(TypedBuilder)]` on a type, you can specify a `#[builder(...)]`
 /// attribute on the type, and on any fields in it.
 ///
 /// On the **type**, the following values are permitted:
@@ -74,21 +64,29 @@ mod util;
 ///   but it won't be a link. If you turn this on, the builder type and its `build` method will get
 ///   sane defaults. The field methods on the builder will be undocumented by default.
 ///
+/// - `crate_module_path`: This is only needed when `typed_builder` is reexported from another
+///   crate - which usually happens when another macro uses it. In that case, it is the
+///   reponsibility of that macro to set the `crate_module_path` to the _unquoted_ module path from
+///   which the `typed_builder` crate can be accessed, so that the `TypedBuilder` macro will be
+///   able to access the typed declared in it.
+///
+///   Defaults to `#[builder(crate_module_path=::typed_builder)]`.
+///
 /// - The following subsections:
 ///   - `builder_method(...)`: customize the builder method that creates the builder type
 ///   - `builder_type(...)`: customize the builder type
 ///   - `build_method(...)`: customize the final build method
 ///
 ///   All have the same fields:
-///   - `vis = "…"`: sets the visibility of the build method, default is `pub`
-///   - `name = …`: sets the fn name of the build method, default is `build`
-///   - `doc = "…"` replaces the default documentation that will be generated for the
+///   - `vis = "..."`: sets the visibility of the build method, default is `pub`
+///   - `name = ...`: sets the fn name of the build method, default is `build`
+///   - `doc = "..."` replaces the default documentation that will be generated for the
 ///     `build()` method of the builder type. Setting this implies `doc`.
 ///
 ///
 /// - The `build_method(...)` subsection also has:
 ///   - `into` or `into = ...`: change the output type of the builder. When a specific value/type
-///     is set via the assignement, this will be the output type of the builder. If no specific
+///     is set via the assignment, this will be the output type of the builder. If no specific
 ///     type is set, but `into` is specified, the return type will be generic and the user can
 ///     decide which type shall be constructed. In both cases an [`Into`] conversion is required to
 ///     be defined from the original type to the target type.
@@ -124,22 +122,34 @@ mod util;
 ///    struct Point { x: f32, y: f32 }
 ///    ```
 ///
+/// - `mutators(...)` takes functions, that can mutate fields inside of the builder.
+///   See [mutators](#mutators) for details.
+///
 /// On each **field**, the following values are permitted:
 ///
 /// - `default`: make the field optional, defaulting to `Default::default()`. This requires that
 ///   the field type implement `Default`. Mutually exclusive with any other form of default.
 ///
-/// - `default = …`: make the field optional, defaulting to the expression `…`.
+/// - `default = ...`: make the field optional, defaulting to the expression `...`.
 ///
-/// - `default_code = "…"`: make the field optional, defaulting to the expression `…`. Note that
+/// - `default_code = "..."`: make the field optional, defaulting to the expression `...`. Note that
 ///    you need to enclose it in quotes, which allows you to use it together with other custom
 ///    derive proc-macro crates that complain about "expected literal".
 ///    Note that if `...` contains a string, you can use raw string literals to avoid escaping the
 ///    double quotes - e.g. `#[builder(default_code = r#""default text".to_owned()"#)]`.
 ///
+/// - `via_mutators`: initialize the field when constructing the builder, useful in combination
+///   with [mutators](#mutators).
+///
+/// - `via_mutators = ...` or `via_mutators(init = ...)`: initialies the field with the expression `...`
+///   when constructing the builder, useful in combination with [mutators](#mutators).
+///
+/// - `mutators(...)` takes functions, that can mutate fields inside of the builder.
+///   Mutators specified on a field, mark this field as required, see [mutators](#mutators) for details.
+///
 /// - `setter(...)`: settings for the field setters. The following values are permitted inside:
 ///
-///   - `doc = "…"`: sets the documentation for the field's setter on the builder type. This will be
+///   - `doc = "..."`: sets the documentation for the field's setter on the builder type. This will be
 ///     of no value unless you enable docs for the builder type with `#[builder(doc)]` or similar on
 ///     the type.
 ///
@@ -162,49 +172,83 @@ mod util;
 ///     `param1: Type1, param2: Type2 ...` instead of the field type itself. The parameters are
 ///     transformed into the field type using the expression `expr`. The transformation is performed
 ///     when the setter is called.
-#[proc_macro_derive(TypedBuilder, attributes(builder))]
-pub fn derive_typed_builder(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
-    let input = parse_macro_input!(input as DeriveInput);
-    match impl_my_derive(&input) {
-        Ok(output) => output.into(),
-        Err(error) => error.to_compile_error().into(),
+///
+///   - `prefix = "..."` prepends the setter method with the specified prefix. For example, setting
+///     `prefix = "with_"` results in setters like `with_x` or `with_y`. This option is combinable
+///     with `suffix = "..."`.
+///
+///   - `suffix = "..."` appends the setter method with the specified suffix. For example, setting
+///     `suffix = "_value"` results in setters like `x_value` or `y_value`. This option is combinable
+///     with `prefix = "..."`.
+///
+///   - `mutable_during_default_resolution`: when expressions in `default = ...` field attributes
+///     are evaluated, this field will be mutable, allowing earlier-defined fields to be mutated by
+///     later-defined fields.
+///     **Warning** - Use this feature with care! If the field that mutates the previous field in
+///     its `default` expression is set via a setter, that mutation will not happen.
+///
+/// # Mutators
+/// Set fields can be mutated using mutators, these can be defind via `mutators(...)`.
+///
+/// Fields annotated with `#[builder(via_mutators)]` are always available to mutators. Additional fields,
+/// that the mutator accesses need to be delcared using `#[mutator(requires = [field1, field2, ...])]`.
+/// The mutator will only be availible to call when they are set.
+///
+/// Mutators on a field, result in them automatically making the field required, i.e., it needs to be
+/// marked as `via_mutators`, or its setter be called. Appart from that, they behave identically.
+///
+/// ```
+/// use typed_builder::TypedBuilder;
+///
+/// #[derive(PartialEq, Debug, TypedBuilder)]
+/// #[builder(mutators(
+///     // Mutator has only acces to fields marked as `via_mutators`.
+///     fn inc_a(&mut self, a: i32){
+///         self.a += a;
+///     }
+///     // Mutator has access to `x` additionally.
+///     #[mutator(requires = [x])]
+///     fn x_into_b(&mut self) {
+///         self.b.push(self.x)
+///     }
+/// ))]
+/// struct Struct {
+///     // Does not require explicit `requires = [x]`, as the field
+///     // the mutator is specifed on, is required implicitly.
+///     #[builder(mutators(
+///         fn x_into_b_field(self) {
+///             self.b.push(self.x)
+///         }
+///     ))]
+///     x: i32,
+///     #[builder(via_mutators(init = 1))]
+///     a: i32,
+///     #[builder(via_mutators)]
+///     b: Vec<i32>
+/// }
+///
+/// // Mutators do not enforce only being called once
+/// assert_eq!(
+///     Struct::builder().x(2).x_into_b().x_into_b().x_into_b_field().inc_a(2).build(),
+///     Struct {x: 2, a: 3, b: vec![2, 2, 2]});
+/// ```
+pub use typed_builder_macro::TypedBuilder;
+
+#[doc(hidden)]
+pub trait Optional<T> {
+    fn into_value<F: FnOnce() -> T>(self, default: F) -> T;
+}
+
+impl<T> Optional<T> for () {
+    fn into_value<F: FnOnce() -> T>(self, default: F) -> T {
+        default()
     }
 }
 
-fn impl_my_derive(ast: &syn::DeriveInput) -> Result<TokenStream, Error> {
-    let data = match &ast.data {
-        syn::Data::Struct(data) => match &data.fields {
-            syn::Fields::Named(fields) => {
-                let struct_info = struct_info::StructInfo::new(ast, fields.named.iter())?;
-                let builder_creation = struct_info.builder_creation_impl()?;
-                let conversion_helper = struct_info.conversion_helper_impl();
-                let fields = struct_info
-                    .included_fields()
-                    .map(|f| struct_info.field_impl(f))
-                    .collect::<Result<Vec<_>, _>>()?;
-                let fields = quote!(#(#fields)*).into_iter();
-                let required_fields = struct_info
-                    .included_fields()
-                    .filter(|f| f.builder_attr.default.is_none())
-                    .map(|f| struct_info.required_field_impl(f))
-                    .collect::<Vec<_>>();
-                let build_method = struct_info.build_method_impl();
-
-                quote! {
-                    #builder_creation
-                    #conversion_helper
-                    #( #fields )*
-                    #( #required_fields )*
-                    #build_method
-                }
-            }
-            syn::Fields::Unnamed(_) => return Err(Error::new(ast.span(), "TypedBuilder is not supported for tuple structs")),
-            syn::Fields::Unit => return Err(Error::new(ast.span(), "TypedBuilder is not supported for unit structs")),
-        },
-        syn::Data::Enum(_) => return Err(Error::new(ast.span(), "TypedBuilder is not supported for enums")),
-        syn::Data::Union(_) => return Err(Error::new(ast.span(), "TypedBuilder is not supported for unions")),
-    };
-    Ok(data)
+impl<T> Optional<T> for (T,) {
+    fn into_value<F: FnOnce() -> T>(self, _: F) -> T {
+        self.0
+    }
 }
 
 // It'd be nice for the compilation tests to live in tests/ with the rest, but short of pulling in
@@ -287,4 +331,20 @@ fn impl_my_derive(ast: &syn::DeriveInput) -> Result<TokenStream, Error> {
 ///
 /// let _ = Foo::builder().x(Uncloneable).clone();
 /// ```
+///
+/// Handling deprecated fields:
+///
+/// ```compile_fail
+/// use typed_builder::TypedBuilder;
+///
+/// #[derive(TypedBuilder)]
+/// struct Foo {
+///     #[deprecated = "Don't use this!"]
+///     #[allow(dead_code)]
+///     value: i32,
+/// }
+///
+/// #[deny(deprecated)]
+/// Foo::builder().value(42).build();
+///```
 fn _compile_fail_tests() {}
